@@ -114,7 +114,7 @@ xb_machine_add_operator (XbMachine *self, const gchar *str, const gchar *name)
  * xb_machine_add_method:
  * @self: a #XbMachine
  * @name: function name, e.g. `contains`
- * @n_opcodes: minimum number of opcodes requried on the stack
+ * @n_opcodes: minimum number of opcodes required on the stack
  * @method_cb: function to call
  * @user_data: user pointer to pass to @method_cb, or %NULL
  * @user_data_free: a function which gets called to free @user_data, or %NULL
@@ -236,10 +236,8 @@ XbOpcode *
 xb_machine_opcode_func_new (XbMachine *self, const gchar *func_name)
 {
 	XbMachineMethodItem *item = xb_machine_find_func (self, func_name);
-	if (item == NULL) {
-		g_critical ("failed to find %s", func_name);
+	if (item == NULL)
 		return NULL;
-	}
 	return xb_opcode_new (XB_OPCODE_KIND_FUNCTION,
 			      g_strdup (func_name),
 			      item->idx, g_free);
@@ -251,25 +249,27 @@ xb_machine_parse_add_func (XbMachine *self,
 			   const gchar *func_name,
 			   GError **error)
 {
-	XbMachineMethodItem *item;
-	XbOpcode *op;
+	XbMachinePrivate *priv = GET_PRIVATE (self);
+	g_autoptr(XbOpcode) opcode = NULL;
 
-	/* find function by name */
-	item = xb_machine_find_func (self, func_name);
-	if (item == NULL) {
+	/* match opcode, which should always exist */
+	opcode = xb_machine_opcode_func_new (self, func_name);
+	if (opcode == NULL) {
 		g_set_error (error,
 			     G_IO_ERROR,
 			     G_IO_ERROR_NOT_SUPPORTED,
-			     "function %s() is not supported",
-			     func_name);
+			     "built-in function not found: %s", func_name);
+		return FALSE;
+	}
+	if (!xb_stack_push_steal (opcodes, g_steal_pointer (&opcode))) {
+		g_set_error (error,
+			     G_IO_ERROR,
+			     G_IO_ERROR_NOT_SUPPORTED,
+			     "stack size %u exhausted",
+			     priv->stack_size);
 		return FALSE;
 	}
 
-	/* create new opcode */
-	op = xb_opcode_new (XB_OPCODE_KIND_FUNCTION,
-			    g_strdup (func_name),
-			    item->idx, g_free);
-	xb_stack_push_steal (opcodes, op);
 	return TRUE;
 }
 
@@ -340,27 +340,30 @@ g_ascii_string_to_unsigned (const gchar *str,
 #endif
 
 static gboolean
-xb_machine_parse_add_text_raw (XbMachine *self,
-			       XbStack *opcodes,
-			       const gchar *str,
-			       GError **error)
+xb_machine_parse_add_text (XbMachine *self,
+			   XbStack *opcodes,
+			   const gchar *text,
+			   gssize text_len,
+			   GError **error)
 {
 	XbMachinePrivate *priv = GET_PRIVATE (self);
+	g_autofree gchar *str = NULL;
 	guint64 val;
-	gsize text_len;
 
 	/* NULL is perfectly valid */
-	if (str == NULL) {
+	if (text == NULL) {
 		xb_stack_push_steal (opcodes, xb_opcode_text_new_static (str));
 		return TRUE;
 	}
 
 	/* never add empty literals */
-	text_len = strlen (str);
+	if (text_len < 0)
+		text_len = strlen (text);
 	if (text_len == 0)
 		return TRUE;
 
 	/* do any additional handlers */
+	str = g_strndup (text, text_len);
 	for (guint i = 0; i < priv->text_handlers->len; i++) {
 		XbMachineTextHandlerItem *item = g_ptr_array_index (priv->text_handlers, i);
 		gboolean handled = FALSE;
@@ -408,82 +411,114 @@ xb_machine_parse_add_text_raw (XbMachine *self,
 		     G_IO_ERROR_NOT_SUPPORTED,
 		     "cannot parse text or number `%s`", str);
 	return FALSE;
+
 }
 
 static gboolean
-xb_machine_parse_add_text (XbMachine *self,
-			   XbStack *opcodes,
-			   const gchar *str,
-			   GError **error)
-{
-	return xb_machine_parse_add_text_raw (self, opcodes, str, error);
-}
-
-static gssize
 xb_machine_parse_section (XbMachine *self,
 			  XbStack *opcodes,
 			  const gchar *text,
-			  gsize start,
 			  gssize text_len,
-			  guint level,
+			  gboolean is_method,
 			  GError **error)
 {
 	XbMachinePrivate *priv = GET_PRIVATE (self);
-	g_autoptr(GString) acc = g_string_new (NULL);
 
-	/* sanity check */
-	if (level > 20) {
-		g_autofree gchar *tmp = g_strndup (text, text_len);
-		g_set_error (error,
-			     G_IO_ERROR,
-			     G_IO_ERROR_INVALID_DATA,
-			     "nesting deeper than 20 levels supported: %s",
-			     tmp);
-		return G_MAXSSIZE;
-	}
+	/* fall back for simplicity */
+	if (text_len < 0)
+		text_len = strlen (text);
+	if (text_len == 0)
+		return TRUE;
 
-	/* build accumulator until hitting either bracket, then recurse */
-	for (gssize i = start; i < text_len; i++) {
-		if (priv->debug_flags & XB_MACHINE_DEBUG_FLAG_SHOW_PARSING)
-			g_debug ("@%u, >%c", level, text[i]);
-		if (text[i] == ',')
-			continue;
-		if (text[i] == '(') {
-			i = xb_machine_parse_section (self, opcodes, text, i + 1, text_len, level + 1, error);
-			if (i == 0) {
-				g_set_error_literal (error,
-						     G_IO_ERROR,
-						     G_IO_ERROR_INVALID_DATA,
-						     "failed to find matching bracket");
-				return G_MAXSSIZE;
+//	g_debug ("xb_machine_parse_section{%s} method=%i", g_strndup (text, text_len), is_method);
+	for (gssize i = 0; i < text_len; i++) {
+		for (guint j = 0; j < priv->operators->len; j++) {
+			XbMachineOperator *op = g_ptr_array_index (priv->operators, j);
+			if (strncmp (text + i, op->str, op->strsz) == 0) {
+				if (is_method) {
+					/* after then before */
+					if (!xb_machine_parse_section (self, opcodes, text + i + op->strsz, -1, is_method, error))
+						return FALSE;
+					if (i > 0) {
+						if (!xb_machine_parse_section (self, opcodes, text, i, FALSE, error))
+							return FALSE;
+					}
+					if (!xb_machine_parse_add_func (self, opcodes, op->name, error))
+						return FALSE;
+				} else {
+					/* before then after */
+					if (i > 0) {
+						if (!xb_machine_parse_section (self, opcodes, text, i, FALSE, error))
+							return FALSE;
+					}
+					if (!xb_machine_parse_section (self, opcodes, text + i + op->strsz, -1, is_method, error))
+						return FALSE;
+					if (!xb_machine_parse_add_func (self, opcodes, op->name, error))
+						return FALSE;
+				}
+				return TRUE;
 			}
-			if (i == G_MAXSSIZE)
-				return G_MAXSSIZE;
-			if (!xb_machine_parse_add_func (self, opcodes, acc->str, error))
-				return G_MAXSSIZE;
-			g_string_truncate (acc, 0);
-			continue;
 		}
-		if (text[i] == ')') {
-			if (acc->len > 0) {
-				if (!xb_machine_parse_add_text (self, opcodes, acc->str, error))
-					return G_MAXSSIZE;
+	}
+
+	/* nothing matched */
+	if (is_method)
+		return xb_machine_parse_add_func (self, opcodes, text, error);
+	return xb_machine_parse_add_text (self, opcodes, text, text_len, error);
+}
+
+static gboolean
+xb_machine_parse_sections (XbMachine *self,
+			   XbStack *opcodes,
+			   const gchar *text,
+			   gsize text_len,
+			   gboolean is_method,
+			   GError **error)
+{
+	g_autofree gchar *tmp = NULL;
+	if (text_len == 0)
+		return TRUE;
+
+	/* leading comma */
+	if (text[0] == ',') {
+		tmp = g_strndup (text + 1, text_len - 1);
+	} else {
+		tmp = g_strndup (text, text_len);
+	}
+
+//	g_debug ("xb_machine_parse_sections{%s} method=%i", tmp, is_method);
+	for (gint i = text_len - 1; i >= 0; i--) {
+//		g_debug ("%u\t\t%c", i, tmp[i]);
+		if (tmp[i] == ',') {
+			tmp[i] = '\0';
+			if (is_method) {
+				if (!xb_machine_parse_add_func (self,
+								opcodes,
+								tmp + i + 1,
+								error))
+					return FALSE;
+				is_method = FALSE;
+			} else {
+				if (!xb_machine_parse_section (self,
+							       opcodes,
+							       tmp + i + 1,
+							       -1,
+							       TRUE,
+							       error))
+					return FALSE;
 			}
-			g_string_truncate (acc, 0);
-			return i;
 		}
-
-		g_string_append_c (acc, text[i]);
 	}
-
-	/* any left over */
-	if (acc->len > 0) {
-		if (!xb_machine_parse_add_text (self, opcodes, acc->str, error))
-			return G_MAXSSIZE;
-		g_string_truncate (acc, 0);
+	if (tmp[0] != '\0') {
+		if (!xb_machine_parse_section (self,
+					       opcodes,
+					       tmp,
+					       -1,
+					       is_method,
+					       error))
+			return FALSE;
 	}
-
-	return 0;
+	return TRUE;
 }
 
 static gchar *
@@ -500,22 +535,6 @@ xb_machine_get_opcodes_sig (XbMachine *self, XbStack *opcodes)
 	return g_string_free (str, FALSE);
 }
 
-static gboolean
-xb_machine_parse_part (XbMachine *self,
-		       XbStack *opcodes,
-		       const gchar *text,
-		       gsize start,
-		       gssize text_len,
-		       GError **error)
-{
-	gsize rc = xb_machine_parse_section (self, opcodes, text, start, text_len, 0, error);
-	if (rc == G_MAXSSIZE) {
-		g_prefix_error (error, "failed to parse part: ");
-		return FALSE;
-	}
-	return TRUE;
-}
-
 static gboolean
 xb_machine_opcodes_optimize_fn (XbMachine *self,
 				XbOpcode *op,
@@ -526,10 +545,9 @@ xb_machine_opcodes_optimize_fn (XbMachine *self,
 {
 	XbMachineMethodItem *item;
 	XbMachinePrivate *priv = GET_PRIVATE (self);
-	XbOpcode *op_tmp;
-	gboolean result = TRUE;
 	g_autofree gchar *stack_str = NULL;
 	g_autoptr(GError) error_local = NULL;
+	g_autoptr(XbOpcode) op_result = NULL;
 	g_autoptr(XbStack) stack = NULL;
 
 	/* a function! lets check the arg length */
@@ -551,13 +569,13 @@ xb_machine_opcodes_optimize_fn (XbMachine *self,
 	/* make a copy of the stack with the arguments */
 	stack = xb_stack_new (item->n_opcodes);
 	for (guint i = item->n_opcodes; i > 0; i--) {
-		op_tmp = g_ptr_array_index (src, *idx - (i + 1));
+		XbOpcode *op_tmp = g_ptr_array_index (src, *idx - (i + 1));
 		xb_stack_push (stack, op_tmp);
 	}
 
 	/* run the method */
 	stack_str = xb_stack_to_string (stack);
-	if (!item->method_cb (self, stack, &result, item->user_data, NULL, &error_local)) {
+	if (!item->method_cb (self, stack, NULL, item->user_data, NULL, &error_local)) {
 		if (priv->debug_flags & XB_MACHINE_DEBUG_FLAG_SHOW_OPTIMIZER) {
 			g_debug ("ignoring opimized call to %s(%s): %s",
 				 item->name,
@@ -569,23 +587,31 @@ xb_machine_opcodes_optimize_fn (XbMachine *self,
 	}
 
 	/* the method ran, add the result and discard the arguments */
-	op_tmp = xb_stack_pop (stack);
-	if (op_tmp != NULL) {
+	op_result = xb_stack_pop (stack);
+	if (op_result == NULL) {
+		g_set_error_literal (error,
+				     G_IO_ERROR,
+				     G_IO_ERROR_INVALID_DATA,
+				     "internal error; no retval on stack");
+		return FALSE;
+	}
+	if (xb_opcode_get_kind (op_result) != XB_OPCODE_KIND_BOOLEAN) {
 		if (priv->debug_flags & XB_MACHINE_DEBUG_FLAG_SHOW_OPTIMIZER)
 			g_debug ("method ran, adding result");
 		*idx -= item->n_opcodes;
-		g_ptr_array_add (dst, op_tmp);
+		g_ptr_array_add (dst, g_steal_pointer (&op_result));
 		return TRUE;
 	}
 
 	/* nothing was added to the stack, so check if the predicate will
 	 * always evaluate to TRUE */
 	if (priv->debug_flags & XB_MACHINE_DEBUG_FLAG_SHOW_OPTIMIZER) {
-		g_debug ("method ran, result %s",
-			 result ? "TRUE" : "FALSE");
+		g_autofree gchar *tmp = xb_opcode_to_string (op_result);
+		g_debug ("method ran, result %s", tmp);
 	}
-	if (result) {
+	if (xb_opcode_get_val (op_result) == TRUE) {
 		*idx -= item->n_opcodes;
+		g_ptr_array_add (dst, g_steal_pointer (&op_result));
 		return TRUE;
 	}
 
@@ -634,6 +660,82 @@ xb_machine_opcodes_optimize (XbMachine *self, XbStack *opcodes, GError **error)
 	return TRUE;
 }
 
+static gsize
+xb_machine_parse_text (XbMachine *self,
+		       XbStack *opcodes,
+		       const gchar *text,
+		       gsize text_len,
+		       guint level,
+		       GError **error)
+{
+	XbMachinePrivate *priv = GET_PRIVATE (self);
+	guint tail = 0;
+
+	/* sanity check */
+	if (level > 20) {
+		g_autofree gchar *tmp = g_strndup (text, text_len);
+		g_set_error (error,
+			     G_IO_ERROR,
+			     G_IO_ERROR_INVALID_DATA,
+			     "nesting deeper than 20 levels supported: %s", tmp);
+		return G_MAXSIZE;
+	}
+
+	//g_debug ("%u xb_machine_parse_text{%s}", level, g_strndup (text, text_len));
+	for (guint i = 0; i < text_len; i++) {
+		if (priv->debug_flags & XB_MACHINE_DEBUG_FLAG_SHOW_PARSING)
+			g_debug ("LVL %u\t%u:\t\t%c", level, i, text[i]);
+		if (text[i] == '(') {
+			gsize j = 0;
+			j = xb_machine_parse_text (self,
+						   opcodes,
+						   text + i + 1,
+						   text_len - i,
+						   level + 1,
+						   error);
+			if (j == G_MAXSIZE)
+				return G_MAXSIZE;
+			if (!xb_machine_parse_sections (self,
+							opcodes,
+							text + tail,
+							i - tail,
+							TRUE,
+							error))
+				return G_MAXSIZE;
+			i += j;
+			tail = i + 1;
+			continue;
+		}
+		if (text[i] == ')') {
+			if (!xb_machine_parse_sections (self,
+							opcodes,
+							text + tail,
+							i - tail,
+							FALSE,
+							error))
+				return G_MAXSIZE;
+			return i + 1;
+		}
+	}
+	if (tail != text_len && level > 0) {
+		g_autofree gchar *tmp = g_strndup (text, text_len);
+		g_set_error (error,
+			     G_IO_ERROR,
+			     G_IO_ERROR_INVALID_DATA,
+			     "brackets did not match: %s", tmp);
+		return G_MAXSIZE;
+	}
+	if (!xb_machine_parse_sections (self,
+					opcodes,
+					text + tail,
+					text_len - tail,
+					FALSE,
+					error))
+		return G_MAXSIZE;
+	return 0;
+}
+
+
 /**
  * xb_machine_parse_full:
  * @self: a #XbMachine
@@ -659,6 +761,7 @@ xb_machine_parse_full (XbMachine *self,
 {
 	XbMachineOpcodeFixupItem *item;
 	XbMachinePrivate *priv = GET_PRIVATE (self);
+	guint level = 0;
 	g_autoptr(XbStack) opcodes = NULL;
 	g_autofree gchar *opcodes_sig = NULL;
 
@@ -677,52 +780,10 @@ xb_machine_parse_full (XbMachine *self,
 		return NULL;
 	}
 
-	/* look for foo=bar */
+	/* parse into opcodes */
 	opcodes = xb_stack_new (priv->stack_size);
-	for (gssize i = 0; i < text_len && xb_stack_get_size (opcodes) == 0; i++) {
-		for (guint j = 0; j < priv->operators->len; j++) {
-			XbMachineOperator *op = g_ptr_array_index (priv->operators, j);
-			if (strncmp (text + i, op->str, op->strsz) == 0) {
-				g_autoptr(XbOpcode) opcode = NULL;
-
-				/* match opcode, which should always exist */
-				opcode = xb_machine_opcode_func_new (self, op->name);
-				if (opcode == NULL) {
-					g_set_error (error,
-						     G_IO_ERROR,
-						     G_IO_ERROR_NOT_SUPPORTED,
-						     "built-in function not found: %s",
-						     op->name);
-					return NULL;
-				}
-				if (!xb_machine_parse_part (self, opcodes, text,
-							    0, /* start */
-							    i, /* end */
-							    error))
-					return NULL;
-				if (!xb_machine_parse_part (self, opcodes, text,
-							    i + op->strsz,
-							    text_len,
-							    error))
-					return NULL;
-				if (!xb_stack_push_steal (opcodes, g_steal_pointer (&opcode))) {
-					g_set_error (error,
-						     G_IO_ERROR,
-						     G_IO_ERROR_NOT_SUPPORTED,
-						     "stack size %u exhausted",
-						     priv->stack_size);
-					return NULL;
-				}
-				break;
-			}
-		}
-	}
-
-	/* remainder */
-	if (xb_stack_get_size (opcodes) == 0) {
-		if (!xb_machine_parse_part (self, opcodes, text, 0, text_len, error))
-			return NULL;
-	}
+	if (xb_machine_parse_text (self, opcodes, text, text_len, level, error) == G_MAXSIZE)
+		return FALSE;
 
 	/* do any fixups */
 	opcodes_sig = xb_machine_get_opcodes_sig (self, opcodes);
@@ -789,7 +850,6 @@ static gboolean
 xb_machine_run_func (XbMachine *self,
 		     XbStack *stack,
 		     XbOpcode *opcode,
-		     gboolean *result,
 		     gpointer exec_data,
 		     GError **error)
 {
@@ -812,7 +872,7 @@ xb_machine_run_func (XbMachine *self,
 			     item->n_opcodes, xb_stack_get_size (stack));
 		return FALSE;
 	}
-	if (!item->method_cb (self, stack, result, item->user_data, exec_data, error)) {
+	if (!item->method_cb (self, stack, NULL, item->user_data, exec_data, error)) {
 		g_prefix_error (error, "failed to call %s(): ", item->name);
 		return FALSE;
 	}
@@ -878,6 +938,7 @@ xb_machine_run (XbMachine *self,
 		GError **error)
 {
 	XbMachinePrivate *priv = GET_PRIVATE (self);
+	g_autoptr(XbOpcode) opcode_success = NULL;
 	g_autoptr(XbStack) stack = NULL;
 
 	g_return_val_if_fail (XB_IS_MACHINE (self), FALSE);
@@ -896,17 +957,15 @@ xb_machine_run (XbMachine *self,
 			if (!xb_machine_run_func (self,
 						  stack,
 						  opcode,
-						  result,
 						  exec_data,
 						  error))
 				return FALSE;
-			if (*result == FALSE)
-				return TRUE;
 			continue;
 		}
 
 		/* add to stack */
 		if (kind == XB_OPCODE_KIND_TEXT ||
+		    kind == XB_OPCODE_KIND_BOOLEAN ||
 		    kind == XB_OPCODE_KIND_INTEGER ||
 		    kind == XB_OPCODE_KIND_INDEXED_TEXT ||
 		    kind == XB_OPCODE_KIND_BOUND_TEXT ||
@@ -936,8 +995,8 @@ xb_machine_run (XbMachine *self,
 		return FALSE;
 	}
 
-	/* the stack should have been completely consumed */
-	if (xb_stack_get_size (stack) > 0) {
+	/* the stack should have one boolean left on the stack */
+	if (xb_stack_get_size (stack) != 1) {
 		g_autofree gchar *tmp = xb_stack_to_string (stack);
 		g_set_error (error,
 			     G_IO_ERROR,
@@ -946,6 +1005,17 @@ xb_machine_run (XbMachine *self,
 			     xb_stack_get_size (stack), tmp);
 		return FALSE;
 	}
+	opcode_success = xb_stack_pop (stack);
+	if (xb_opcode_get_kind (opcode_success) != XB_OPCODE_KIND_BOOLEAN) {
+		g_autofree gchar *tmp = xb_stack_to_string (stack);
+		g_set_error (error,
+			     G_IO_ERROR,
+			     G_IO_ERROR_INVALID_DATA,
+			     "Expected boolean, got: %s", tmp);
+		return FALSE;
+	}
+	if (result != NULL)
+		*result = xb_opcode_get_val (opcode_success);
 
 	/* success */
 	return TRUE;
@@ -1088,7 +1158,7 @@ xb_machine_stack_push_text_steal (XbMachine *self, XbStack *stack, gchar *str)
  * xb_machine_stack_push_integer:
  * @self: a #XbMachine
  * @stack: a #XbStack
- * @val: interger literal
+ * @val: integer literal
  *
  * Adds an integer literal to the stack.
  *
@@ -1108,7 +1178,7 @@ xb_machine_stack_push_integer (XbMachine *self, XbStack *stack, guint32 val)
 /**
  * xb_machine_set_stack_size:
  * @self: a #XbMachine
- * @stack_size: interger
+ * @stack_size: integer
  *
  * Sets the maximum stack size used for the machine.
  *
@@ -1144,6 +1214,60 @@ xb_machine_get_stack_size (XbMachine *self)
 	return priv->stack_size;
 }
 
+static gboolean
+xb_machine_func_and_cb (XbMachine *self,
+			XbStack *stack,
+			gboolean *result,
+			gpointer user_data,
+			gpointer exec_data,
+			GError **error)
+{
+	g_autoptr(XbOpcode) op1 = xb_machine_stack_pop (self, stack);
+	g_autoptr(XbOpcode) op2 = xb_machine_stack_pop (self, stack);
+
+	/* INTE:INTE */
+	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
+		xb_stack_push_bool (stack, xb_opcode_get_val (op1) && xb_opcode_get_val (op2));
+		return TRUE;
+	}
+
+	/* fail */
+	g_set_error (error,
+		     G_IO_ERROR,
+		     G_IO_ERROR_NOT_SUPPORTED,
+		     "%s:%s types not supported",
+		     xb_opcode_kind_to_string (xb_opcode_get_kind (op1)),
+		     xb_opcode_kind_to_string (xb_opcode_get_kind (op2)));
+	return FALSE;
+}
+
+static gboolean
+xb_machine_func_or_cb (XbMachine *self,
+		       XbStack *stack,
+		       gboolean *result,
+		       gpointer user_data,
+		       gpointer exec_data,
+		       GError **error)
+{
+	g_autoptr(XbOpcode) op1 = xb_machine_stack_pop (self, stack);
+	g_autoptr(XbOpcode) op2 = xb_machine_stack_pop (self, stack);
+
+	/* INTE:INTE */
+	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
+		xb_stack_push_bool (stack, xb_opcode_get_val (op1) || xb_opcode_get_val (op2));
+		return TRUE;
+	}
+
+	/* fail */
+	g_set_error (error,
+		     G_IO_ERROR,
+		     G_IO_ERROR_NOT_SUPPORTED,
+		     "%s:%s types not supported",
+		     xb_opcode_kind_to_string (xb_opcode_get_kind (op1)),
+		     xb_opcode_kind_to_string (xb_opcode_get_kind (op2)));
+	return FALSE;
+}
+
 static gboolean
 xb_machine_func_eq_cb (XbMachine *self,
 		       XbStack *stack,
@@ -1157,14 +1281,14 @@ xb_machine_func_eq_cb (XbMachine *self,
 
 	/* INTE:INTE */
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
-		*result = xb_opcode_get_val (op1) == xb_opcode_get_val (op2);
+		xb_stack_push_bool (stack, xb_opcode_get_val (op1) == xb_opcode_get_val (op2));
 		return TRUE;
 	}
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_strcmp0 (xb_opcode_get_str (op1),
-				     xb_opcode_get_str (op2)) == 0;
+		xb_stack_push_bool (stack, g_strcmp0 (xb_opcode_get_str (op1),
+						      xb_opcode_get_str (op2)) == 0);
 		return TRUE;
 	}
 
@@ -1172,7 +1296,7 @@ xb_machine_func_eq_cb (XbMachine *self,
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_str (op2)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op2) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op2),
@@ -1180,7 +1304,7 @@ xb_machine_func_eq_cb (XbMachine *self,
 						 &val, error)) {
 			return FALSE;
 		}
-		*result = val == xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, val == xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
@@ -1188,7 +1312,7 @@ xb_machine_func_eq_cb (XbMachine *self,
 	if (xb_opcode_cmp_val (op2) && xb_opcode_cmp_str (op1)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op1) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op1),
@@ -1196,7 +1320,7 @@ xb_machine_func_eq_cb (XbMachine *self,
 						 &val, error)) {
 			return FALSE;
 		}
-		*result = val == xb_opcode_get_val (op2);
+		xb_stack_push_bool (stack, val == xb_opcode_get_val (op2));
 		return TRUE;
 	}
 
@@ -1223,14 +1347,14 @@ xb_machine_func_ne_cb (XbMachine *self,
 
 	/* INTE:INTE */
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
-		*result = xb_opcode_get_val (op1) != xb_opcode_get_val (op2);
+		xb_stack_push_bool (stack, xb_opcode_get_val (op1) != xb_opcode_get_val (op2));
 		return TRUE;
 	}
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_strcmp0 (xb_opcode_get_str (op1),
-				     xb_opcode_get_str (op2)) != 0;
+		xb_stack_push_bool (stack, g_strcmp0 (xb_opcode_get_str (op1),
+						      xb_opcode_get_str (op2)) != 0);
 		return TRUE;
 	}
 
@@ -1238,7 +1362,7 @@ xb_machine_func_ne_cb (XbMachine *self,
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_str (op2)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op2) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op2),
@@ -1246,7 +1370,7 @@ xb_machine_func_ne_cb (XbMachine *self,
 						 &val, error)) {
 			return FALSE;
 		}
-		*result = val != xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, val != xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
@@ -1273,14 +1397,14 @@ xb_machine_func_lt_cb (XbMachine *self,
 
 	/* INTE:INTE */
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
-		*result = xb_opcode_get_val (op2) < xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, xb_opcode_get_val (op2) < xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_strcmp0 (xb_opcode_get_str (op2),
-				     xb_opcode_get_str (op1)) < 0;
+		xb_stack_push_bool (stack, g_strcmp0 (xb_opcode_get_str (op2),
+						      xb_opcode_get_str (op1)) < 0);
 		return TRUE;
 	}
 
@@ -1288,7 +1412,7 @@ xb_machine_func_lt_cb (XbMachine *self,
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_str (op2)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op2) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op2),
@@ -1296,7 +1420,7 @@ xb_machine_func_lt_cb (XbMachine *self,
 						 &val, error)) {
 			return FALSE;
 		}
-		*result = val < xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, val < xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
@@ -1323,14 +1447,14 @@ xb_machine_func_gt_cb (XbMachine *self,
 
 	/* INTE:INTE */
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
-		*result = xb_opcode_get_val (op2) > xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, xb_opcode_get_val (op2) > xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_strcmp0 (xb_opcode_get_str (op2),
-				     xb_opcode_get_str (op1)) > 0;
+		xb_stack_push_bool (stack, g_strcmp0 (xb_opcode_get_str (op2),
+						      xb_opcode_get_str (op1)) > 0);
 		return TRUE;
 	}
 
@@ -1338,7 +1462,7 @@ xb_machine_func_gt_cb (XbMachine *self,
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_str (op2)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op2) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op2),
@@ -1346,7 +1470,7 @@ xb_machine_func_gt_cb (XbMachine *self,
 						 &val, error)) {
 			return FALSE;
 		}
-		*result = val > xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, val > xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
@@ -1373,14 +1497,14 @@ xb_machine_func_le_cb (XbMachine *self,
 
 	/* INTE:INTE */
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
-		*result = xb_opcode_get_val (op2) <= xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, xb_opcode_get_val (op2) <= xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_strcmp0 (xb_opcode_get_str (op2),
-				     xb_opcode_get_str (op1)) <= 0;
+		xb_stack_push_bool (stack, g_strcmp0 (xb_opcode_get_str (op2),
+						      xb_opcode_get_str (op1)) <= 0);
 		return TRUE;
 	}
 
@@ -1388,7 +1512,7 @@ xb_machine_func_le_cb (XbMachine *self,
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_str (op2)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op2) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op2),
@@ -1396,7 +1520,7 @@ xb_machine_func_le_cb (XbMachine *self,
 						 &val, error)) {
 			return FALSE;
 		}
-		*result = val <= xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, val <= xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
@@ -1474,13 +1598,13 @@ xb_machine_func_not_cb (XbMachine *self,
 
 	/* TEXT */
 	if (xb_opcode_cmp_str (op1)) {
-		*result = xb_opcode_get_str (op1) == NULL;
+		xb_stack_push_bool (stack, xb_opcode_get_str (op1) == NULL);
 		return TRUE;
 	}
 
 	/* INTE */
 	if (xb_opcode_cmp_val (op1)) {
-		*result = xb_opcode_get_val (op1) == 0;
+		xb_stack_push_bool (stack, xb_opcode_get_val (op1) == 0);
 		return TRUE;
 	}
 
@@ -1506,14 +1630,14 @@ xb_machine_func_ge_cb (XbMachine *self,
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_strcmp0 (xb_opcode_get_str (op2),
-				     xb_opcode_get_str (op1)) >= 0;
+		xb_stack_push_bool (stack, g_strcmp0 (xb_opcode_get_str (op2),
+						      xb_opcode_get_str (op1)) >= 0);
 		return TRUE;
 	}
 
 	/* INTE:INTE */
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_val (op2)) {
-		*result = xb_opcode_get_val (op2) >= xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, xb_opcode_get_val (op2) >= xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
@@ -1521,7 +1645,7 @@ xb_machine_func_ge_cb (XbMachine *self,
 	if (xb_opcode_cmp_val (op1) && xb_opcode_cmp_str (op2)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op2) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op2),
@@ -1529,7 +1653,7 @@ xb_machine_func_ge_cb (XbMachine *self,
 						 &val, error)) {
 			return FALSE;
 		}
-		*result = val >= xb_opcode_get_val (op1);
+		xb_stack_push_bool (stack, val >= xb_opcode_get_val (op1));
 		return TRUE;
 	}
 
@@ -1556,8 +1680,8 @@ xb_machine_func_contains_cb (XbMachine *self,
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = xb_string_contains (xb_opcode_get_str (op2),
-					      xb_opcode_get_str (op1));
+		xb_stack_push_bool (stack, xb_string_contains (xb_opcode_get_str (op2),
+							       xb_opcode_get_str (op1)));
 		return TRUE;
 	}
 
@@ -1584,8 +1708,8 @@ xb_machine_func_starts_with_cb (XbMachine *self,
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_str_has_prefix (xb_opcode_get_str (op2),
-					    xb_opcode_get_str (op1));
+		xb_stack_push_bool (stack, g_str_has_prefix (xb_opcode_get_str (op2),
+							     xb_opcode_get_str (op1)));
 		return TRUE;
 	}
 
@@ -1612,8 +1736,8 @@ xb_machine_func_ends_with_cb (XbMachine *self,
 
 	/* TEXT:TEXT */
 	if (xb_opcode_cmp_str (op1) && xb_opcode_cmp_str (op2)) {
-		*result = g_str_has_suffix (xb_opcode_get_str (op2),
-					    xb_opcode_get_str (op1));
+		xb_stack_push_bool (stack, g_str_has_suffix (xb_opcode_get_str (op2),
+							     xb_opcode_get_str (op1)));
 		return TRUE;
 	}
 
@@ -1641,7 +1765,7 @@ xb_machine_func_number_cb (XbMachine *self,
 	if (xb_opcode_cmp_str (op1)) {
 		guint64 val = 0;
 		if (xb_opcode_get_str (op1) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		if (!g_ascii_string_to_unsigned (xb_opcode_get_str (op1),
@@ -1675,7 +1799,7 @@ xb_machine_func_strlen_cb (XbMachine *self,
 	/* TEXT */
 	if (xb_opcode_cmp_str (op1)) {
 		if (xb_opcode_get_str (op1) == NULL) {
-			*result = FALSE;
+			xb_stack_push_bool (stack, FALSE);
 			return TRUE;
 		}
 		xb_machine_stack_push_integer (self, stack, strlen (xb_opcode_get_str (op1)));
@@ -1762,7 +1886,9 @@ xb_machine_init (XbMachine *self)
 	priv->opcode_fixup = g_hash_table_new_full (g_str_hash, g_str_equal,
 						     g_free, (GDestroyNotify) xb_machine_opcode_fixup_free);
 
-	/* build-in functions */
+	/* built-in functions */
+	xb_machine_add_method (self, "and", 2, xb_machine_func_and_cb, NULL, NULL);
+	xb_machine_add_method (self, "or", 2, xb_machine_func_or_cb, NULL, NULL);
 	xb_machine_add_method (self, "eq", 2, xb_machine_func_eq_cb, NULL, NULL);
 	xb_machine_add_method (self, "ne", 2, xb_machine_func_ne_cb, NULL, NULL);
 	xb_machine_add_method (self, "lt", 2, xb_machine_func_lt_cb, NULL, NULL);
@@ -1780,6 +1906,10 @@ xb_machine_init (XbMachine *self)
 	xb_machine_add_method (self, "string-length", 1, xb_machine_func_strlen_cb, NULL, NULL);
 
 	/* built-in operators */
+	xb_machine_add_operator (self, " and ", "and");
+	xb_machine_add_operator (self, " or ", "or");
+	xb_machine_add_operator (self, "&&", "and");
+	xb_machine_add_operator (self, "||", "or");
 	xb_machine_add_operator (self, "!=", "ne");
 	xb_machine_add_operator (self, "<=", "le");
 	xb_machine_add_operator (self, ">=", "ge");
